
///////////////////////////////////////////////////////////////////////////////
//
//	isd_edit.cpp
//

#include <vector>
#include "isd.h"
#include "project.h"
#include "id.h"
#include "undo.h"


class ISDEdit_UndoMethod : public UndoMethod{

	struct ISDChange{
		int index;
		LevelISD data_before;
		LevelISD data_after;
		ISDChange(int _i, const LevelISD &_bef, const LevelISD &_aft)
			: index(_i), data_before(_bef), data_after(_aft){}
	};
	typedef std::vector<ISDChange> ISDVector;

private:
	ISDManager *isd;
	ISDVector data;

private:
	LPCSTR text_undo() const{ return "Xe[WX^C̕ύX"; }
	LPCSTR text_redo() const{ return "Xe[WX^C̕ύX"; }
	void undo(){
		for (ISDVector::iterator i=data.begin(); i!=data.end(); i++){
			isd->data[(*i).index] = (*i).data_before;
		}
		isd->save();
		isd->project->levels.redraw_all();
	}
	void redo(){
		for (ISDVector::iterator i=data.begin(); i!=data.end(); i++){
			isd->data[(*i).index] = (*i).data_after;
		}
		isd->save();
		isd->project->levels.redraw_all();
	}

public:
	ISDEdit_UndoMethod(ISDManager *_isd) : isd(_isd){}
	~ISDEdit_UndoMethod(){}
	void add_change(int index, const LevelISD &before, const LevelISD &after){
		data.push_back(ISDChange(index, before, after));
	}

};





class ISDEditDialog : public MyDialog, public RomData_Base{
	ISDManager *isd;
	LevelISD *data;
	bool *change;
	int isel;
	bool edit_change_enabled;
private:
	LPCSTR get_template() const{ return "ISD_EDITDLG"; }
	BOOL Proc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam){
		switch (msg){
			case WM_INITDIALOG:
				init((WORD)lParam);
				return TRUE;
			case WM_COMMAND:
				switch (LOWORD(wParam)){
					case IDC_EDIT0: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT0); return TRUE; } break;
					case IDC_EDIT1: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT1); return TRUE; } break;
					case IDC_EDIT2: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT2); return TRUE; } break;
					case IDC_EDIT3: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT3); return TRUE; } break;
					case IDC_EDIT5: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT5); return TRUE; } break;
					case IDC_EDIT6: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT6); return TRUE; } break;
					case IDC_EDIT7: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT7); return TRUE; } break;
					case IDC_EDIT8: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT8); return TRUE; } break;
					case IDC_EDIT9: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDIT9); return TRUE; } break;
					case IDC_EDITA: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDITA); return TRUE; } break;
					case IDC_EDITB: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDITB); return TRUE; } break;
					case IDC_EDITC: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDITC); return TRUE; } break;
					case IDC_EDITD: if (HIWORD(wParam) == EN_CHANGE){ edit_changed(IDC_EDITD); return TRUE; } break;
					case IDC_CHECK0: case IDC_CHECK1: case IDC_CHECK2: case IDC_CHECK3:
					case IDC_CHECK4: case IDC_CHECK5: case IDC_CHECK6: case IDC_CHECK7:
						check_changed();
						return TRUE;
					case IDC_LIST0:
						if (HIWORD(wParam) == LBN_SELCHANGE){
							sel_changed(SendDlgItemMessage(hWnd, IDC_LIST0, LB_GETCURSEL, 0, 0));
							return TRUE;
						}
						break;
					case IDOK:
						save_data();
						EndDialog(hDlg, IDOK);
						return TRUE;
					case IDCANCEL:
						EndDialog(hDlg, IDCANCEL);
						return TRUE;
				}
				break;
		}
		return FALSE;
	}
	void init(WORD style_init){
		SendDlgItemMessage(hWnd, IDC_EDIT0, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT1, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT2, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT3, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT5, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT6, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT7, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT8, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT9, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDITA, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDITB, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDITC, EM_SETLIMITTEXT, 2, 0);
		SendDlgItemMessage(hWnd, IDC_EDITD, EM_SETLIMITTEXT, 2, 0);
		HWND hwStyleList = GetDlgItem(hWnd, IDC_LIST0);
		char s[256];
		int style_count = isd->get_style_count();
		WORD style_begin = isd->get_style_begin();
		WORD style = style_begin;
		for (int i=0; i<style_count; i++){
			sprintf_s(s, sizeof s, "%04X: %s", style, isd->style_name[i]);
			SendMessage(hwStyleList, LB_INSERTSTRING, i, (LPARAM)s);
			style += 21;
		}
		int isel_init = (style_init - style_begin) / 21;
		if (isel_init < 0 || isel_init >= style_count) isel_init = 0;
		SendMessage(hwStyleList, LB_SETCURSEL, isel_init, 0);
		sel_changed(isel_init);
	}
	void sel_changed(int isel_new){
		edit_change_enabled = false;
		isel = isel_new;
		WORD style = isd->get_style_begin() + isel*21;
		char s[32];
		sprintf_s(s, sizeof s, "%02X", data[isel].palette1);
		SetDlgItemText(hWnd, IDC_EDIT0, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].effect);
		SetDlgItemText(hWnd, IDC_EDIT1, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].music);
		SetDlgItemText(hWnd, IDC_EDIT2, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].palette2);
		SetDlgItemText(hWnd, IDC_EDIT3, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].pointer1);
		SetDlgItemText(hWnd, IDC_EDIT5, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].pointer2);
		SetDlgItemText(hWnd, IDC_EDIT6, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].fg_scroll);
		SetDlgItemText(hWnd, IDC_EDIT7, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].tilemap_graph);
		SetDlgItemText(hWnd, IDC_EDIT8, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].redraw);
		SetDlgItemText(hWnd, IDC_EDIT9, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].tilemap_scroll);
		SetDlgItemText(hWnd, IDC_EDITA, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].map_num);
		SetDlgItemText(hWnd, IDC_EDITB, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].view_thg);
		SetDlgItemText(hWnd, IDC_EDITC, s);
		sprintf_s(s, sizeof s, "%02X", data[isel].special_action);
		SetDlgItemText(hWnd, IDC_EDITD, s);
		CheckDlgButton(hWnd, IDC_CHECK0, (data[isel].special_action & 0x01) != 0);
		CheckDlgButton(hWnd, IDC_CHECK1, (data[isel].special_action & 0x02) != 0);
		CheckDlgButton(hWnd, IDC_CHECK2, (data[isel].special_action & 0x04) != 0);
		CheckDlgButton(hWnd, IDC_CHECK3, (data[isel].special_action & 0x08) != 0);
		CheckDlgButton(hWnd, IDC_CHECK4, (data[isel].special_action & 0x10) != 0);
		CheckDlgButton(hWnd, IDC_CHECK5, (data[isel].special_action & 0x20) != 0);
		CheckDlgButton(hWnd, IDC_CHECK6, (data[isel].special_action & 0x40) != 0);
		CheckDlgButton(hWnd, IDC_CHECK7, (data[isel].special_action & 0x80) != 0);
		edit_change_enabled = true;
	}
	void style_changed(){
		if (change[isel]) return;
		change[isel] = true;
		char s[256];
		sprintf_s(s, sizeof s, "*%04X: %s", isd->style_begin + isel*LevelISD::SIZE, isd->style_name[isel]);
		HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
		SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)isel, 0);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)isel, (LPARAM)s);
		SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
		SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
		return;
	}
	void edit_changed(WORD id){
		if (!edit_change_enabled) return;
		char s[64];
		int i;
		switch (id){
			case IDC_EDIT0:
				GetDlgItemText(hWnd, IDC_EDIT0, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].palette1 = (BYTE)i;
				break;
			case IDC_EDIT1:
				GetDlgItemText(hWnd, IDC_EDIT1, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].effect = (BYTE)i;
				break;
			case IDC_EDIT2:
				GetDlgItemText(hWnd, IDC_EDIT2, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].music = (BYTE)i;
				break;
			case IDC_EDIT3:
				GetDlgItemText(hWnd, IDC_EDIT3, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].palette2 = (WORD)i;
				break;
			case IDC_EDIT5:
				GetDlgItemText(hWnd, IDC_EDIT5, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].pointer1 = (WORD)i;
				break;
			case IDC_EDIT6:
				GetDlgItemText(hWnd, IDC_EDIT6, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].pointer2 = (WORD)i;
				break;
			case IDC_EDIT7:
				GetDlgItemText(hWnd, IDC_EDIT7, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].fg_scroll = (BYTE)i;
				break;
			case IDC_EDIT8:
				GetDlgItemText(hWnd, IDC_EDIT8, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].tilemap_graph = (BYTE)i;
				break;
			case IDC_EDIT9:
				GetDlgItemText(hWnd, IDC_EDIT9, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].redraw = (BYTE)i;
				break;
			case IDC_EDITA:
				GetDlgItemText(hWnd, IDC_EDITA, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].tilemap_scroll = (BYTE)i;
				break;
			case IDC_EDITB:
				GetDlgItemText(hWnd, IDC_EDITB, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].map_num = (BYTE)i;
				break;
			case IDC_EDITC:
				GetDlgItemText(hWnd, IDC_EDITC, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].view_thg = (BYTE)i;
				break;
			case IDC_EDITD:
				GetDlgItemText(hWnd, IDC_EDITD, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].special_action = (BYTE)i;
				CheckDlgButton(hWnd, IDC_CHECK0, (data[isel].special_action & 0x01) != 0);
				CheckDlgButton(hWnd, IDC_CHECK1, (data[isel].special_action & 0x02) != 0);
				CheckDlgButton(hWnd, IDC_CHECK2, (data[isel].special_action & 0x04) != 0);
				CheckDlgButton(hWnd, IDC_CHECK3, (data[isel].special_action & 0x08) != 0);
				CheckDlgButton(hWnd, IDC_CHECK4, (data[isel].special_action & 0x10) != 0);
				CheckDlgButton(hWnd, IDC_CHECK5, (data[isel].special_action & 0x20) != 0);
				CheckDlgButton(hWnd, IDC_CHECK6, (data[isel].special_action & 0x40) != 0);
				CheckDlgButton(hWnd, IDC_CHECK7, (data[isel].special_action & 0x80) != 0);
				break;
		}
		style_changed();
	}
	void check_changed(){
		char s[256];
		edit_change_enabled = false;
		data[isel].special_action = 0x00;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK0) != 0) << 0;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK1) != 0) << 1;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK2) != 0) << 2;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK3) != 0) << 3;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK4) != 0) << 4;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK5) != 0) << 5;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK6) != 0) << 6;
		data[isel].special_action |= (IsDlgButtonChecked(hWnd, IDC_CHECK7) != 0) << 7;
		sprintf_s(s, sizeof s, "%02X", data[isel].special_action);
		SetDlgItemText(hWnd, IDC_EDITD, s);
		edit_change_enabled = true;
		style_changed();
	}
	void save_data(){
		int n=0;
		ISDEdit_UndoMethod *undo = new ISDEdit_UndoMethod(isd);
		for (int i=0; i<isd->style_count; i++){
			undo->add_change(i, isd->data[i], data[i]);
			if (change[i]){
				isd->data[i] = data[i];
				n++;
			}
		}
		if (n){
			isd->save();
			isd->project->undo_buffer.add_method(undo);
		}else{
			delete undo;
		}
	}
public:
	ISDEditDialog(ISDManager *_isd) : isd(_isd), data(NULL), change(NULL){
		data = new LevelISD[isd->style_count];
		change = new bool[isd->style_count];
		for (int i=0; i<isd->style_count; i++){
			data[i] = isd->data[i];
			change[i] = false;
		}
	}
	~ISDEditDialog(){
		if (data) delete[] data;
	}
};




int ISDManager::style_edit_dialog(HWND hw_owner, WORD init_style){
	ISDEditDialog dialog(this);
	return dialog.dialog_box(hw_owner, (LPARAM)init_style);
}



