
///////////////////////////////////////////////////////////////////////////////
//
//	level_bananas.h
//

#ifndef _LEVEL_BANANAS_
#define _LEVEL_BANANAS_

///////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <list>
#include "base.h"


class LevelBananasManager;
class LevelManager;
class ProjectManager;


//	BananaAllocData: oiizuf[^(\Ă)
//---------------------------------------------------------
struct BananaAllocData{
	static const int size = 4;
	int pdx;
	int pdy;
	bool fore;
	bool fore2;
	BYTE type;
	void load_data(const ProjectManager *project, BANK bank, SADRS adrs, WORD code);
	void store_data(ProjectManager *project, BANK bank, SADRS adrs, WORD code) const;
	void normalize(WORD code);
};


//	LevelBananasData: Xe[W̃oiizuf[^
//---------------------------------------------------------
class LevelBananasData : public RomData_Base{
	std::list<BananaAllocData> bananas;
public:
	typedef std::list<BananaAllocData>::iterator iterator;
	typedef std::list<BananaAllocData>::reverse_iterator reverse_iterator;
	void load_data(const ProjectManager *project, BANK bank, SADRS adrs, int level_num, WORD code);
	void store_data(ProjectManager *project, BANK bank, SADRS adrs, int level_num, WORD code) const;
	void clear();
	iterator begin(){ return bananas.begin(); }
	iterator end(){ return bananas.end(); }
	reverse_iterator rbegin(){ return bananas.rbegin(); }
	reverse_iterator rend(){ return bananas.rend(); }
	size_t get_banana_count(){ return bananas.size(); }
};



//---------------------------------------------------------
class LevelBananaApply_UndoMethod : public UndoMethod, public RomData_Base{
	ProjectManager *project_manager;
	LevelManager *level_manager;
	LevelBananasManager *bananas_manager;
	BananaAllocData data_before;
	BananaAllocData data_after;
	int level_num;
	int banana_num;
	virtual LPCSTR text_undo() const{ return "oii̕ύX"; }
	virtual LPCSTR text_redo() const{ return "oii̕ύX"; }
	virtual void undo();
	virtual void redo();
	void method_in_level(bool bUndo);
	void method_out_of_level(bool bUndo);
public:
	LevelBananaApply_UndoMethod(ProjectManager *_project_manager, LevelManager *_level_manager,
		LevelBananasManager *_bananas_manager, int _level_num, int _banana_num,
		const BananaAllocData *_data_before, const BananaAllocData *_data_after);
	~LevelBananaApply_UndoMethod();
};

//---------------------------------------------------------
class LevelBananaUpDown_UndoMethod : public UndoMethod, public RomData_Base{
	void method_in_level(bool bUp);
	void method_out_of_level(bool bUp);
protected:
	ProjectManager *project_manager;
	LevelManager *level_manager;
	LevelBananasManager *bananas_manager;
	int level_num;
	int banana_former;
	int banana_latter;
	LevelBananaUpDown_UndoMethod(ProjectManager *_project_manager, LevelManager *_level_manager,
		LevelBananasManager *_bananas_manager, int _level_num, int _banana_former);
	~LevelBananaUpDown_UndoMethod();
	void method(bool bUp);
};

class LevelBananaUp_UndoMethod : public LevelBananaUpDown_UndoMethod{
	virtual LPCSTR text_undo() const{ return "oii"; }
	virtual LPCSTR text_redo() const{ return "oii"; }
	virtual void undo();
	virtual void redo();
public:
	LevelBananaUp_UndoMethod(ProjectManager *_project_manager, LevelManager *_level_manager,
		LevelBananasManager *_bananas_manager, int _level_num, int _banana_num);
	~LevelBananaUp_UndoMethod();
};

class LevelBananaDown_UndoMethod : public LevelBananaUpDown_UndoMethod{
	virtual LPCSTR text_undo() const{ return "oii"; }
	virtual LPCSTR text_redo() const{ return "oii"; }
	virtual void undo();
	virtual void redo();
public:
	LevelBananaDown_UndoMethod(ProjectManager *_project_manager, LevelManager *_level_manager,
		LevelBananasManager *_bananas_manager, int _level_num, int _banana_num);
	~LevelBananaDown_UndoMethod();
};

class LevelBananaSort_UndoMethod : public UndoMethod, public RomData_Base{
	ProjectManager *project_manager;
	LevelManager *level_manager;
	LevelBananasManager *bananas_manager;
	int level_num;
	LevelBananasData data_before;
	LevelBananasData data_after;
	int banana_before;
	int banana_after;
	virtual LPCSTR text_undo() const{ return "oiĩ\[g"; }
	virtual LPCSTR text_redo() const{ return "oiĩ\[g"; }
	virtual void undo();
	virtual void redo();
	void method_in_level(bool bUndo);
	void method_out_of_level(bool bUndo);
public:
	LevelBananaSort_UndoMethod(ProjectManager *_project_manager, LevelManager *_level_manager,
		LevelBananasManager *_bananas_manager, int _level_num, int _banana_before, int _banana_after,
		const LevelBananasData &_data_before, const LevelBananasData &_data_after);
	~LevelBananaSort_UndoMethod();
};



//	BananaSetWindow: oiiZbg`悷EBhE
//---------------------------------------------------------
class BananaSetWindow : public RomData_Base, public Window{
	ProjectManager *project;
	HFONT hfText;
public:
	BananaSetWindow(ProjectManager *_project);
	~BananaSetWindow();
	void set_window(HWND _hWnd);
	void clear();
	void output(BYTE type);
};


//	LevelBananasManager: Xe[W̃oiizuǗ
//---------------------------------------------------------
class LevelBananasManager : public RomData_Base, public MyDialog, public TabPage{
	friend class LevelBananaApply_UndoMethod;
	friend class LevelBananaUpDown_UndoMethod;
	friend class LevelBananaSort_UndoMethod;

private:

	virtual LPCSTR get_text() const{ return "oii"; }
	virtual LPCSTR get_template() const{ return "LEVEL_BANANAS"; }
	virtual BOOL Proc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);

	HIMAGELIST hilBananaImage;
	int banana_image_index;
	static const int BANANA_WIDTH = 24;
	static const int BANANA_HEIGHT = 36;
	HPEN	hpnBananaLine;
	HPEN	hpnBananaSel;
	HBRUSH	hbrBananaRect;
	HBRUSH	hbrBananaLine;
	HBRUSH	hbrBananaSel;
	HFONT	hftBananaText;
	enum ShowBananas{
		BAN_HIDE=0,
		BAN_NORMAL,
		BAN_SIMPLE
	} iShowBananas;
	int isel;
	bool bDragging;
	int dragging_x;
	int dragging_y;
	int drag_temp_x;
	int drag_temp_y;
	bool unable_edit_change;

	int level_num;
	LevelBananasData data;
	BananaSetWindow banana_set;

	LevelManager *level;
	ProjectManager *project;

	void set_menuitem_state();
	void sel_changed(int sel);
	void jump(int sel);
	void focus_on_banana(int sel);
	void ban_apply(int sel);
	void ban_up(int sel);
	void ban_down(int sel);
	void ban_sort(int sel);
	void get_banana_text(char *dst, size_t size, const BananaAllocData *banana);
	LevelBananasData::iterator get_iterator(int iIndex);
	void input_banana_data(BananaAllocData *banana);
	void output_banana_data(const BananaAllocData *banana, int iIndex);


public:

	LevelBananasManager(LevelManager *_level, ProjectManager *_project);
	~LevelBananasManager();
	virtual BOOL create(HWND hwTab);
	BOOL create_sys();
	void save_data();
	void load(int _level_num);
	void clear();
	void draw(BitmapBuffer *buf, const RECT *rtRedraw);
	void draw_f(BitmapBuffer *buf, const RECT *rtRedraw);
	BOOL command(WPARAM wParam, LPARAM lParam);
	BOOL translate_dialog_message(MSG *msg);
	void set_state(bool bEnabled);
	bool mouseover(int pdx, int pdy);
	bool accepts_click(int pdx, int pdy);
	void dragging(int pdx, int pdy);
	void release_click(bool bApply);
	void release_sel();

	LADRS get_banana_levels_adrs() const{
		DWORD dw_adrs = GetPrivateProfileInt("Bananas", "Levels", 0x3DAC21, RomInfoFileName);
		LADRS ladrs;
		ladrs.bank = (BANK)((dw_adrs&0xff0000)>>16);
		ladrs.adrs = (SADRS)(dw_adrs&0xffff);
		return ladrs;
	}

};





///////////////////////////////////////////////////////////////////////////////

#endif // _LEVEL_BANANAS_



