
///////////////////////////////////////////////////////////////////////////////
//
//	maps_edit.cpp
//

#include <vector>
#include "maps.h"
#include "project.h"
#include "id.h"



class MapEdit_UndoMethod : public UndoMethod{

	struct MapChange{
		int map_num;
		MapData data_before;
		MapData data_after;
		MapChange(int _i, const MapData &_bef, const MapData &_aft)
			: map_num(_i), data_before(_bef), data_after(_aft){}
	};
	typedef std::vector<MapChange> MapVector;

private:
	MapsManager *maps;
	MapVector data;

private:
	LPCSTR text_undo() const{ return "}bv̏ڍאݒ̕ύX"; }
	LPCSTR text_redo() const{ return "}bv̏ڍאݒ̕ύX"; }
	void undo(){
		for (MapVector::iterator i=data.begin(); i!=data.end(); i++){
			maps->data[(*i).map_num] = (*i).data_before;
		}
		maps->save();
		maps->project->levels.redraw_all();
	}
	void redo(){
		for (MapVector::iterator i=data.begin(); i!=data.end(); i++){
			maps->data[(*i).map_num] = (*i).data_after;
		}
		maps->save();
		maps->project->levels.redraw_all();
	}

public:
	MapEdit_UndoMethod(MapsManager *_maps) : maps(_maps){}
	~MapEdit_UndoMethod(){}
	void add_change(int map_num, const MapData &before, const MapData &after){
		data.push_back(MapChange(map_num, before, after));
	}

};




class MapEditDialog : public MyDialog, public RomData_Base{
	MapsManager *maps;
	MapData *data;
	bool *change;
	int isel;
	bool edit_change_enabled;
private:
	LPCSTR get_template() const{ return "MAP_EDITDLG"; }
	BOOL Proc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam){
		switch (msg){
			case WM_INITDIALOG:
				init((int)lParam);
				return TRUE;
			case WM_COMMAND:
				switch (LOWORD(wParam)){
					case IDC_EDIT0: case IDC_EDIT1: case IDC_EDIT2:
					case IDC_EDIT3: case IDC_EDIT4: case IDC_EDIT5:
					case IDC_EDIT6: case IDC_EDIT7: case IDC_EDIT8:
						if (HIWORD(wParam) == EN_CHANGE){
							edit_changed(LOWORD(wParam));
						}
						break;
					case IDC_LIST0:
						if (HIWORD(wParam) == LBN_SELCHANGE){
							sel_changed(SendDlgItemMessage(hWnd, IDC_LIST0, LB_GETCURSEL, 0, 0));
							return TRUE;
						}
						break;
					case IDOK:
						save_data();
						EndDialog(hDlg, IDOK);
						return TRUE;
					case IDCANCEL:
						EndDialog(hDlg, IDCANCEL);
						return TRUE;
				}
				break;
		}
		return FALSE;
	}
	void init(int map_num){
		SendDlgItemMessage(hWnd, IDC_EDIT0, EM_SETLIMITTEXT, 6, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT1, EM_SETLIMITTEXT, 6, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT2, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT3, EM_SETLIMITTEXT, 6, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT4, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT5, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT6, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT7, EM_SETLIMITTEXT, 4, 0);
		SendDlgItemMessage(hWnd, IDC_EDIT8, EM_SETLIMITTEXT, 4, 0);
		char s[256];
		HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
		for (int i=0; i<maps->map_count; i++){
			sprintf_s(s, sizeof s, "%02X: %s", i, maps->map_name[i]);
			SendMessage(hwList, LB_INSERTSTRING, i, (LPARAM)s);
		}
		SendMessage(hwList, LB_SETCURSEL, map_num, 0);
		sel_changed(map_num);
	}
	void sel_changed(int isel_new){
		isel = isel_new;
		char s[32];
		edit_change_enabled = false;
		sprintf_s(s, sizeof s, "%06X", data[isel].tilemap_adrs.get());
		SetDlgItemText(hWnd, IDC_EDIT0, s);
		sprintf_s(s, sizeof s, "%06X", data[isel].chip_method_adrs.get());
		SetDlgItemText(hWnd, IDC_EDIT1, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].layer_info);
		SetDlgItemText(hWnd, IDC_EDIT2, s);
		sprintf_s(s, sizeof s, "%06X", data[isel].action_map.get());
		SetDlgItemText(hWnd, IDC_EDIT3, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].action_map_size);
		SetDlgItemText(hWnd, IDC_EDIT4, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].special_action_adrs);
		SetDlgItemText(hWnd, IDC_EDIT5, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].tilemap_color);
		SetDlgItemText(hWnd, IDC_EDIT6, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].tilemap_type);
		SetDlgItemText(hWnd, IDC_EDIT7, s);
		sprintf_s(s, sizeof s, "%04X", data[isel].allocation_adrs);
		SetDlgItemText(hWnd, IDC_EDIT8, s);
		edit_change_enabled = true;
	}
	void edit_changed(WORD id){
		if (!edit_change_enabled) return;
		char s[64];
		int i;
		switch (id){
			case IDC_EDIT0:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].tilemap_adrs = (DWORD)i;
				break;
			case IDC_EDIT1:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].chip_method_adrs = (DWORD)i;
				break;
			case IDC_EDIT2:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].layer_info = (WORD)i;
				break;
			case IDC_EDIT3:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].action_map = (DWORD)i;
				break;
			case IDC_EDIT4:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].action_map_size = (WORD)i;
				break;
			case IDC_EDIT5:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].special_action_adrs = (SADRS)i;
				break;
			case IDC_EDIT6:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].tilemap_color = (WORD)i;
				break;
			case IDC_EDIT7:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].tilemap_type = (WORD)i;
				break;
			case IDC_EDIT8:
				GetDlgItemText(hWnd, id, s, sizeof s);
				if (sscanf_s(s, "%x", &i) < 0) i = 0;
				data[isel].allocation_adrs = (WORD)i;
				break;
		}
		map_changed();
	}
	void map_changed(){
		if (change[isel]) return;
		change[isel] = true;
		char s[256];
		sprintf_s(s, sizeof s, "*%02X: %s", isel, maps->map_name[isel]);
		HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
		SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)isel, 0);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)isel, (LPARAM)s);
		SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
		SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	}
	void save_data(){
		int n=0;
		MapEdit_UndoMethod *undo = new MapEdit_UndoMethod(maps);
		for (int i=0; i<maps->map_count; i++){
			undo->add_change(i, maps->data[i], data[i]);
			if (change[i]){
				maps->data[i] = data[i];
				n++;
			}
		}
		if (n){
			maps->save();
			maps->project->undo_buffer.add_method(undo);
		}else{
			delete undo;
		}
	}
public:
	MapEditDialog(MapsManager *_maps) : maps(_maps), data(NULL), change(NULL){
		data = new MapData[maps->map_count];
		change = new bool[maps->map_count];
		for (int i=0; i<maps->map_count; i++){
			data[i] = maps->data[i];
			change[i] = false;
		}
	}
	~MapEditDialog(){
		if (data) delete[] data;
		if (change) delete[] change;
	}

};



int MapsManager::map_edit_dialog(HWND hw_owner, int init_map_num){

	MapEditDialog dialog(this);

	return dialog.dialog_box(hw_owner, (LPARAM)init_map_num);
}









