
///////////////////////////////////////////////////////////////////////////////
//
//	maps.cpp
//

#include "maps.h"
#include "project.h"
#include "id.h"


MapsManager::MapsManager(ProjectManager *_project) : project(_project), map_name(NULL), data(NULL){
	map_count = GetPrivateProfileInt("Maps", "Count", 21, RomInfoFileName);
	if (map_count <= 0 || map_count > 256) map_count = 21;
	pTilemapAdrs	= GetPrivateProfileInt("Maps", "TilemapAdrs",	 0x35BAEF, RomInfoFileName);
	pChipMethodAdrs	= GetPrivateProfileInt("Maps", "ChipMethodAdrs", 0x35BB2E, RomInfoFileName);
	pLayerInfo		= GetPrivateProfileInt("Maps", "LayerInfo",		 0x35BB6D, RomInfoFileName);
	pActionMap		= GetPrivateProfileInt("Maps", "ActionMap",		 0x35BB97, RomInfoFileName);
	pActionMapSize	= GetPrivateProfileInt("Maps", "ActionMapSize",	 0x35BBD6, RomInfoFileName);
	pSpActionAdrs	= GetPrivateProfileInt("Maps", "SpActionAdrs",	 0x35BC00, RomInfoFileName);
	pTilemapColor	= GetPrivateProfileInt("Maps", "TilemapColor",	 0x35BC2A, RomInfoFileName);
	pTilemapType	= GetPrivateProfileInt("Maps", "TilemapType",	 0x35BC54, RomInfoFileName);
	pAllocationAdrs	= GetPrivateProfileInt("Maps", "AllocationAdrs", 0x35BC7E, RomInfoFileName);
	map_name = new char[map_count][256];
	char str[8];
	for (int i=0; i<map_count; i++){
		sprintf_s(str, sizeof str, "%02X", i);
		GetPrivateProfileString("MapName", str, "?", map_name[i], 256, RomInfoFileName);
	}
	data = new MapData[map_count];
}

MapsManager::~MapsManager(){
	if (data) delete[] data;
	if (map_name) delete[] map_name;
}


void MapsManager::load(){
	for (int i=0; i<map_count; i++){
		data[i].tilemap_adrs		= project->read_ladrs(pTilemapAdrs.bank, pTilemapAdrs.adrs + i*3);
		if (data[i].tilemap_adrs.bank >= 0xC0)
		{
			data[i].tilemap_adrs.bank -= 0xC0;
		}
		data[i].chip_method_adrs	= project->read_ladrs(pChipMethodAdrs.bank, pChipMethodAdrs.adrs + i*3);
		if (data[i].chip_method_adrs.bank >= 0xC0)
		{
			data[i].chip_method_adrs.bank -= 0xC0;
		}
		data[i].layer_info			= project->read_word(pLayerInfo.bank, pLayerInfo.adrs + i*2);
		data[i].action_map			= project->read_ladrs(pActionMap.bank, pActionMap.adrs + i*3);
		if (data[i].action_map.bank >= 0xC0)
		{
			data[i].action_map.bank -= 0xC0;
		}
		data[i].action_map_size		= project->read_word(pActionMapSize.bank, pActionMapSize.adrs + i*2);
		data[i].special_action_adrs	= project->read_sadrs(pSpActionAdrs.bank, pSpActionAdrs.adrs + i*2);
		data[i].tilemap_color		= project->read_word(pTilemapColor.bank, pTilemapColor.adrs + i*2);
		data[i].tilemap_type		= project->read_word(pTilemapType.bank, pTilemapType.adrs + i*2);
		data[i].allocation_adrs		= project->read_sadrs(pAllocationAdrs.bank, pAllocationAdrs.adrs + i*2);
	}
}

void MapsManager::save(){
	for (int i=0; i<map_count; i++){
		LADRS ladrs = data[i].tilemap_adrs;
		if (ladrs.bank < 0x40)
		{
			if (ladrs.bank == 0 && ladrs.adrs == 0)
			{
				break;
			}
			ladrs.bank += 0xC0;
		}
		project->write_ladrs(ladrs,							pTilemapAdrs.bank, pTilemapAdrs.adrs + i*3);
		ladrs = data[i].chip_method_adrs;
		if (ladrs.bank < 0x40)
		{
			if (ladrs.bank == 0 && ladrs.adrs == 0)
			{
				break;
			}
			ladrs.bank += 0xC0;
		}
		project->write_ladrs(ladrs,							pChipMethodAdrs.bank, pChipMethodAdrs.adrs + i*3);
		project->write_word(data[i].layer_info,				pLayerInfo.bank, pLayerInfo.adrs + i*2);
		ladrs = data[i].action_map;
		if (ladrs.bank < 0x40)
		{
			if (ladrs.bank == 0 && ladrs.adrs == 0)
			{
				break;
			}
			ladrs.bank += 0xC0;
		}
		project->write_ladrs(ladrs,							pActionMap.bank, pActionMap.adrs + i*3);
		project->write_word(data[i].action_map_size,		pActionMapSize.bank, pActionMapSize.adrs + i*2);
		project->write_sadrs(data[i].special_action_adrs,	pSpActionAdrs.bank, pSpActionAdrs.adrs + i*2);
		project->write_word(data[i].tilemap_color,			pTilemapColor.bank, pTilemapColor.adrs + i*2);
		project->write_word(data[i].tilemap_type,			pTilemapType.bank, pTilemapType.adrs + i*2);
		project->write_sadrs(data[i].allocation_adrs,		pAllocationAdrs.bank, pAllocationAdrs.adrs + i*2);
	}
}

const char *MapsManager::get_map_name(int map_num){
	return map_name[map_num];
}

const MapData *MapsManager::get_map_data(int map_num){
	return &data[map_num];
}

// n`f[^ɗp\ȗ̈̃TCY߂
size_t MapsManager::get_tilemap_max_size(int map_num){
	int next_begin = GetPrivateProfileInt("Tilemap", "AvailableEnd", 0x2572DE, RomInfoFileName);
	int this_begin = data[map_num].tilemap_adrs.get();
	int p;
	for (int i=0; i<map_count; i++){
		p = data[i].tilemap_adrs.get();
		if (p > this_begin && p < next_begin){
			next_begin = p;
		}
	}
	return (size_t)(next_begin - this_begin);
}

// n`f[^̑(cʁEʂё)߂
void MapsManager::get_tilemap_attribute(int map_num, bool *bVert, int *width){
	get_tilemap_attribute(&data[map_num], bVert, width);
	return;
}
void MapsManager::get_tilemap_attribute(const MapData *map_data, bool *bVert, int *width){
	WORD primary_code = project->read_word(pAllocationAdrs.bank, map_data->allocation_adrs + 0xC);
	switch (primary_code){
		case 0x0000: *bVert = false; break;
		case 0x0004: *bVert = true; break;
		case 0x0005: *bVert = true; break;
		case 0x0006: *bVert = true; break;
		case 0x0007: *bVert = true; break;
		case 0x0008: *bVert = false; break;
		default: *bVert = false;
	}
	*width = *bVert?
		project->read_word(pAllocationAdrs.bank, map_data->allocation_adrs)/32:		// x_size
		project->read_word(pAllocationAdrs.bank, map_data->allocation_adrs + 2)/32;	// y_size
	return;
}


void MapsManager::get_chip_file(int map_num, char *buf, size_t buf_size){
	char s[8];
	sprintf_s(s, sizeof s, "%02X", map_num);
	GetPrivateProfileString("MapChip", s, "", buf, buf_size, RomInfoFileName);
	return;
}

void MapsManager::get_chip_file_me(int map_num, char *buf, size_t buf_size){
	switch (map_num){
		case 0: strcpy_s(buf, buf_size, "mapchip_00.bmp"); break;
		case 1: strcpy_s(buf, buf_size, "mapchip_01.bmp"); break;
		case 2: strcpy_s(buf, buf_size, "mapchip_02.bmp"); break;
		case 3: strcpy_s(buf, buf_size, "mapchip_03.bmp"); break;
		case 4: strcpy_s(buf, buf_size, "mapchip_04.bmp"); break;
		case 5: strcpy_s(buf, buf_size, "mapchip_05.bmp"); break;
		case 6: strcpy_s(buf, buf_size, "mapchip_06.bmp"); break;
		case 7: strcpy_s(buf, buf_size, "mapchip_02.bmp"); break;
		case 8: strcpy_s(buf, buf_size, "mapchip_08.bmp"); break;
		case 9: strcpy_s(buf, buf_size, "mapchip_09.bmp"); break;
		case 10: strcpy_s(buf, buf_size, "mapchip_10.bmp"); break;
		case 11: strcpy_s(buf, buf_size, "mapchip_11.bmp"); break;
		case 12: strcpy_s(buf, buf_size, "mapchip_12.bmp"); break;
		case 13: strcpy_s(buf, buf_size, "mapchip_13.bmp"); break;
		case 14: strcpy_s(buf, buf_size, "mapchip_14.bmp"); break;
		case 15: strcpy_s(buf, buf_size, "mapchip_15.bmp"); break;
		case 16: strcpy_s(buf, buf_size, ""); break;
		case 17: strcpy_s(buf, buf_size, "mapchip_14.bmp"); break;
		case 18: strcpy_s(buf, buf_size, "mapchip_10.bmp"); break;
		case 19: strcpy_s(buf, buf_size, "mapchip_05.bmp"); break;
		case 20: strcpy_s(buf, buf_size, "mapchip_11.bmp"); break;
	}
	return;
}


BOOL MapsManager::command(WPARAM wParam, LPARAM lParam){
	switch (LOWORD(wParam)){
		case IDM_MAP_TILEMAP:
			{
				LevelSettingData setting;
				project->levels.get_setting_data(&setting);
				if (tilemap_edit_dialog(
					project->get_hWnd(),
					project->isd.get_style_data(setting.style)->map_num,
					setting.map_index
					) == IDOK)
				{
					project->levels.redraw_all();
				}
			}
			return TRUE;
		case IDM_MAP_MAPEDIT:
			{
				LevelSettingData setting;
				project->levels.get_setting_data(&setting);
				if (map_edit_dialog(project->get_hWnd(), project->isd.get_style_data(setting.style)->map_num) == IDOK){
					project->levels.redraw_all();
				}
			}
			return TRUE;
	}
	return FALSE;
}

void MapsManager::set_state(BOOL bEnabled){
	HMENU hmMain = GetMenu(project->get_hWnd());
	UINT uState = (bEnabled ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND;
	EnableMenuItem(hmMain, IDM_MAP_TILEMAP, uState);
	EnableMenuItem(hmMain, IDM_MAP_MAPEDIT, uState);
	SendMessage(project->get_hwTool(), TB_ENABLEBUTTON, (WPARAM)IDM_MAP_TILEMAP, (LPARAM)bEnabled);
	return;
}



