
///////////////////////////////////////////////////////////////////////////////
//
//	tilemap.cpp
//

#include <vector>
#include "maps.h"
#include "project.h"
#include "id.h"

#ifdef _MSC_VER
#pragma comment(lib, "winmm.lib")
#endif


static const int tilemap_format_count = 6;
static char *tilemap_format[] = {
	"kf[^",
	"WJf[^",
	"sdk2_mapedit݊PPJt@C",
	"sdk2_mapedit݊FMFt@C",
	"map_conv݊PPJt@C",
	"map_conv݊FMFt@C",
};
static char *tilemap_format_filter[] = {
	"kn`f[^(*.sdk2map)\0*.sdk2map\0SẴt@C(*.*)\0*.*\0\0",
	"WJn`f[^(*.bin)\0*.bin\0SẴt@C(*.*)\0*.*\0\0",
	"Platinum Project File(*.ppj)\0*.ppj\0SẴt@C(*.*)\0*.*\0\0",
	"Fixed Mapdata File(*.fmf)\0*.fmf\0SẴt@C(*.*)\0*.*\0\0",
	"Platinum Project File(*.ppj)\0*.ppj\0SẴt@C(*.*)\0*.*\0\0",
	"Fixed Mapdata File(*.fmf)\0*.fmf\0SẴt@C(*.*)\0*.*\0\0",
};
static char *tilemap_format_ext[] = {
	"sdk2map",
	"bin",
	"ppj",
	"fmf",
	"ppj",
	"fmf",
};


static const int map_code_count = 6;
static struct {
	WORD code;
	char *name;
} map_code[] = {
	{ 0x0000, "0000: ʏ퉡" },
	{ 0x0004, "0004: ̍Lc" },
	{ 0x0005, "0005: גc" },
	{ 0x0006, "0006: `ɋ߂c" },
	{ 0x0007, "0007: `ɋ߂c" },
	{ 0x0008, "0008: ג" },
};



struct MapAllocationData{
	WORD width;
	WORD height;
	int area_count;
	struct AreaAllocation{
		WORD left;
		WORD right;
		WORD top;
		WORD bottom;
		WORD code;
	} areas[16];
public:
	void load(const ProjectManager *project, BANK bank, SADRS adrs){
		SADRS p = adrs;
		width = project->read_word(bank, p);	p += 2;
		height = project->read_word(bank, p);	p += 2;
		for (int i=0; i<=16; i++){
			if (project->read_word(bank, p) == 0xFFFF || i==16){
				area_count = i;
				break;
			}
			areas[i].left   = project->read_word(bank, p);	p += 2;
			areas[i].right  = project->read_word(bank, p);	p += 2;
			areas[i].top    = project->read_word(bank, p);	p += 2;
			areas[i].bottom = project->read_word(bank, p);	p += 2;
			areas[i].code   = project->read_word(bank, p);	p += 2;
		}
		return;
	}
	void store(ProjectManager *project, BANK bank, SADRS adrs){
		SADRS p = adrs;
		project->write_word(width, bank, p);	p += 2;
		project->write_word(height, bank, p);	p += 2;
		for (int i=0; i<area_count; i++){
			project->write_word(areas[i].left,   bank, p);	p += 2;
			project->write_word(areas[i].right,  bank, p);	p += 2;
			project->write_word(areas[i].top,    bank, p);	p += 2;
			project->write_word(areas[i].bottom, bank, p);	p += 2;
			project->write_word(areas[i].code,   bank, p);	p += 2;
		}
		project->write_word(0xFFFF, bank, p);
		return;
	}
};



class Tilemap_UndoMethod : public UndoMethod{

	struct MapAllocationChange{
		int map_num;
		MapAllocationData data_before;
		MapAllocationData data_after;
		MapAllocationChange(int _map_num, const MapAllocationData &_bef, const MapAllocationData &_aft)
			: map_num(_map_num), data_before(_bef), data_after(_aft)
		{}
	};
	typedef std::vector<MapAllocationChange> AllocChangeVector;

	struct TilemapChange{
		int map_num;
		BYTE *data_before;
		BYTE *data_after;
		DWORD size;
		TilemapChange(int _map_num, const BYTE *_bef, const BYTE *_aft, DWORD _size)
			: map_num(_map_num), size(_size)
		{
			data_before = new BYTE[size];
			data_after = new BYTE[size];
			memcpy_s(data_before, size, _bef, size);
			memcpy_s(data_after, size, _aft, size);
		}
		~TilemapChange(){
			delete[] data_before;
			delete[] data_after;
		}
	};
	typedef std::vector<TilemapChange*> TilemapChangeVector;

private:
	MapsManager *maps;
	AllocChangeVector alloc;
	TilemapChangeVector tilemap;

	LPCSTR text_undo() const{ return "n`f[^ƃXe[Wzu̕ҏW"; }
	LPCSTR text_redo() const{ return "n`f[^ƃXe[Wzu̕ҏW"; }
	void undo(){
		for (AllocChangeVector::iterator i=alloc.begin(); i!=alloc.end(); i++){
			(*i).data_before.store(maps->project, maps->pAllocationAdrs.bank,
				maps->data[(*i).map_num].allocation_adrs);
		}
		for (TilemapChangeVector::iterator i=tilemap.begin(); i!=tilemap.end(); i++){
			maps->project->write_data((const void*)(*i)->data_before, (*i)->size,
				maps->data[(*i)->map_num].tilemap_adrs);
		}
		maps->project->levels.redraw_all();
	}
	void redo(){
		for (AllocChangeVector::iterator i=alloc.begin(); i!=alloc.end(); i++){
			(*i).data_after.store(maps->project, maps->pAllocationAdrs.bank,
				maps->data[(*i).map_num].allocation_adrs);
		}
		for (TilemapChangeVector::iterator i=tilemap.begin(); i!=tilemap.end(); i++){
			maps->project->write_data((const void*)(*i)->data_after, (*i)->size,
				maps->data[(*i)->map_num].tilemap_adrs);
		}
		maps->project->levels.redraw_all();
	}

public:
	Tilemap_UndoMethod(MapsManager *_maps) : maps(_maps){}
	~Tilemap_UndoMethod(){
		for (TilemapChangeVector::iterator i=tilemap.begin(); i!=tilemap.end(); i++){
			delete *i;
		}
	}
	void add_alloc_change(int map_num, const MapAllocationData &bef, const MapAllocationData &aft){
		alloc.push_back(MapAllocationChange(map_num, bef, aft));
	}
	void add_tilemap_change(int map_num, const BYTE *bef, const BYTE *aft, DWORD size){
		tilemap.push_back(new TilemapChange(map_num, bef, aft, size));
	}

};







class TilemapDialog : public MyDialog, public RomData_Base{
	MapsManager *maps;
	MapAllocationData **alloc_change;
	MapAllocationData alloc_present;
	BYTE **tilemap_change;
	DWORD *tilemap_size;
	int isel;
	int edit_change_disabled;
private:
	LPCSTR get_template() const{ return "TILEMAP_EDIT"; }
	BOOL Proc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam){
		switch (msg){
			case WM_INITDIALOG:
				init(LOWORD(lParam), HIWORD(lParam));
				return TRUE;
			case WM_COMMAND:
				switch (LOWORD(wParam)){
					case IDC_EDIT2: case IDC_EDIT3:
						if (HIWORD(wParam) == EN_CHANGE){
							edit_changed(LOWORD(wParam));
							return TRUE;
						}
						break;
					case IDC_EDIT4: case IDC_EDIT5: case IDC_EDIT6: case IDC_EDIT7:
						if (HIWORD(wParam) == EN_CHANGE){
							edit_changed_alloc(LOWORD(wParam));
							return TRUE;
						}
						break;
					case IDC_COMBO1:
						if (HIWORD(wParam) == CBN_SELCHANGE){
							combo_changed_alloc();
							return TRUE;
						}
						break;
					case IDC_LIST0:
						if (HIWORD(wParam) == LBN_SELCHANGE){
							sel_changed(SendDlgItemMessage(hWnd, IDC_LIST0, LB_GETCURSEL, 0, 0));
							return TRUE;
						}
						break;
					case IDC_LIST1:
						if (HIWORD(wParam) == LBN_SELCHANGE){
							sel_changed_alloc(SendDlgItemMessage(hWnd, IDC_LIST1, LB_GETCURSEL, 0, 0));
							return TRUE;
						}
						break;
					case IDC_BUTTON0:
						import_tilemap();
						return TRUE;
					case IDC_BUTTON1:
						export_tilemap();
						return TRUE;
					case IDOK:		end(IDOK);		return TRUE;
					case IDCANCEL:	end(IDCANCEL);	return TRUE;
				}
				break;
		}
		return FALSE;
	}
	void init(int map_num, int init_map_index){
		edit_change_disabled = 0;
		char s[256];
		HWND hwFormat = GetDlgItem(hWnd, IDC_COMBO0);
		for (int i=0; i<tilemap_format_count; i++){
			SendMessage(hwFormat, CB_INSERTSTRING, i, (LPARAM)tilemap_format[i]);
		}
		int format = GetPrivateProfileInt("MapIO", "DefaultFormat", 0, ConfigFileName);
		SendMessage(hwFormat, CB_SETCURSEL, format, 0);
		HWND hwCode = GetDlgItem(hWnd, IDC_COMBO1);
		for (int i=0; i<map_code_count; i++){
			SendMessage(hwCode, CB_INSERTSTRING, i, (LPARAM)map_code[i].name);
		}
		HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
		for (int i=0; i<maps->map_count; i++){
			sprintf_s(s, sizeof s, "%02X: %s", i, maps->map_name[i]);
			SendMessage(hwList, LB_INSERTSTRING, i, (LPARAM)s);
		}
		SendMessage(hwList, LB_SETCURSEL, map_num, 0);
		sel_changed(map_num, init_map_index);
		return;
	}
	void sel_changed(int map_num, int init_map_index=0){
		if (isel == map_num) return;
		isel = map_num;
		edit_change_disabled++;
		char s[256];
		if (maps->data[map_num].tilemap_type & 0x0020){
			// n`f[^
			CheckDlgButton(hWnd, IDC_CHECK0, TRUE);
			EnableWindow(GetDlgItem(hWnd, IDC_CHECK1), TRUE);
			CheckDlgButton(hWnd, IDC_CHECK1, (maps->data[map_num].tilemap_type & 0x0400 ? TRUE : FALSE));
			sprintf_s(s, sizeof s, "%06X", maps->data[map_num].tilemap_adrs.get());
			SetDlgItemText(hWnd, IDC_EDIT0, s);
			sprintf_s(s, sizeof s, "%06X", maps->get_tilemap_max_size(map_num));
			SetDlgItemText(hWnd, IDC_EDIT1, s);
			EnableWindow(GetDlgItem(hWnd, IDC_BUTTON0), TRUE);
			EnableWindow(GetDlgItem(hWnd, IDC_BUTTON1), TRUE);
		}else{
			// n`f[^Ȃ
			CheckDlgButton(hWnd, IDC_CHECK0, FALSE);
			EnableWindow(GetDlgItem(hWnd, IDC_CHECK1), FALSE);
			CheckDlgButton(hWnd, IDC_CHECK1, FALSE);
			SetDlgItemText(hWnd, IDC_EDIT0, "-");
			SetDlgItemText(hWnd, IDC_EDIT1, "-");
			EnableWindow(GetDlgItem(hWnd, IDC_BUTTON0), FALSE);
			EnableWindow(GetDlgItem(hWnd, IDC_BUTTON1), FALSE);
		}
		HWND hwAlloc = GetDlgItem(hWnd, IDC_LIST1);
		SendMessage(hwAlloc, WM_SETREDRAW, FALSE, 0);
		SendMessage(hwAlloc, LB_RESETCONTENT, 0, 0);
		MapAllocationData *alloc_data;
		if (alloc_change[map_num]){
			alloc_data = alloc_change[map_num];
		}else{
			alloc_present.load(maps->project, maps->pAllocationAdrs.bank, maps->data[map_num].allocation_adrs);
			alloc_data = &alloc_present;
		}
		sprintf_s(s, sizeof s, "%d", alloc_data->width);
		SetDlgItemText(hWnd, IDC_EDIT2, s);
		sprintf_s(s, sizeof s, "%d", alloc_data->height);
		SetDlgItemText(hWnd, IDC_EDIT3, s);
		for (int i=0; i<alloc_data->area_count; i++){
			sprintf_s(s, sizeof s, "#%d: [%04X] (%d, %d, %d, %d)", i,
				alloc_data->areas[i].code,
				alloc_data->areas[i].left, alloc_data->areas[i].right,
				alloc_data->areas[i].top, alloc_data->areas[i].bottom
				);
			SendMessage(hwAlloc, LB_INSERTSTRING, i, (LPARAM)s);
		}
		SendMessage(hwAlloc, LB_SETCURSEL, init_map_index, 0);
		SendMessage(hwAlloc, WM_SETREDRAW, TRUE, 0);
		sel_changed_alloc(init_map_index);
		edit_change_disabled--;
		return;
	}
	void sel_changed_alloc(int area){
		char s[256];
		edit_change_disabled++;
		MapAllocationData *alloc_data = alloc_change[isel]? alloc_change[isel] : &alloc_present;
		sprintf_s(s, sizeof s, "%d", alloc_data->areas[area].left);
		SetDlgItemText(hWnd, IDC_EDIT4, s);
		sprintf_s(s, sizeof s, "%d", alloc_data->areas[area].right);
		SetDlgItemText(hWnd, IDC_EDIT5, s);
		sprintf_s(s, sizeof s, "%d", alloc_data->areas[area].top);
		SetDlgItemText(hWnd, IDC_EDIT6, s);
		sprintf_s(s, sizeof s, "%d", alloc_data->areas[area].bottom);
		SetDlgItemText(hWnd, IDC_EDIT7, s);
		int i;
		for (i=0; i<map_code_count; i++){
			if (map_code[i].code == alloc_data->areas[area].code) break;
		}
		if (i == map_code_count){
			MessageBox(hWnd, "W̌n̒lzOłB", "x", MB_OK | MB_ICONEXCLAMATION);
			i = -1;
		}
		SendDlgItemMessage(hWnd, IDC_COMBO1, CB_SETCURSEL, i, 0);
		edit_change_disabled--;
	}
	void data_changed(){
		char s[256];
		sprintf_s(s, sizeof s, "*%02X: %s", isel, maps->map_name[isel]);
		HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
		SendMessage(hwList, WM_SETREDRAW, FALSE, 0);
		SendMessage(hwList, LB_DELETESTRING, isel, 0);
		SendMessage(hwList, LB_INSERTSTRING, isel, (LPARAM)s);
		SendMessage(hwList, LB_SETCURSEL, isel, 0);
		SendMessage(hwList, WM_SETREDRAW, TRUE, 0);
	}
	void alloc_data_change(){
		alloc_change[isel] = new MapAllocationData();
		*alloc_change[isel] = alloc_present;
		data_changed();
		return;
	}
	void edit_changed(WORD id){
		if (edit_change_disabled) return;
		if (!alloc_change[isel]){
			alloc_data_change();
		}
		char s[256];
		int i;
		GetDlgItemText(hWnd, id, s, sizeof s);
		sscanf_s(s, "%d", &i);
		switch (id){
			case IDC_EDIT2: alloc_change[isel]->width  = i; break;
			case IDC_EDIT3: alloc_change[isel]->height = i; break;
		}
		return;
	}
	void edit_changed_alloc(WORD id){
		if (edit_change_disabled) return;
		if (!alloc_change[isel]){
			alloc_data_change();
		}
		int isel_alloc = SendDlgItemMessage(hWnd, IDC_LIST1, LB_GETCURSEL, 0, 0);
		char s[256];
		int i;
		GetDlgItemText(hWnd, id, s, sizeof s);
		sscanf_s(s, "%d", &i);
		switch (id){
			case IDC_EDIT4: alloc_change[isel]->areas[isel_alloc].left   = i; break;
			case IDC_EDIT5: alloc_change[isel]->areas[isel_alloc].right  = i; break;
			case IDC_EDIT6: alloc_change[isel]->areas[isel_alloc].top    = i; break;
			case IDC_EDIT7: alloc_change[isel]->areas[isel_alloc].bottom = i; break;
		}
		alloc_data_changed(isel_alloc);
		return;
	}
	void combo_changed_alloc(){
		if (edit_change_disabled) return;
		if (!alloc_change[isel]){
			alloc_data_change();
		}
		int isel_alloc = SendDlgItemMessage(hWnd, IDC_LIST1, LB_GETCURSEL, 0, 0);
		int code_index = SendDlgItemMessage(hWnd, IDC_COMBO1, CB_GETCURSEL, 0, 0);
		alloc_change[isel]->areas[isel_alloc].code = map_code[code_index].code;
		alloc_data_changed(isel_alloc);
		return;
	}
	void alloc_data_changed(int isel_alloc){
		char s[256];
		sprintf_s(s, sizeof s, "#%d: [%04X] (%d, %d, %d, %d)", isel_alloc,
			alloc_change[isel]->areas[isel_alloc].code,
			alloc_change[isel]->areas[isel_alloc].left, alloc_change[isel]->areas[isel_alloc].right,
			alloc_change[isel]->areas[isel_alloc].top, alloc_change[isel]->areas[isel_alloc].bottom
			);
		HWND hwList = GetDlgItem(hWnd, IDC_LIST1);
		SendMessage(hwList, WM_SETREDRAW, FALSE, 0);
		SendMessage(hwList, LB_DELETESTRING, isel_alloc, 0);
		SendMessage(hwList, LB_INSERTSTRING, isel_alloc, (LPARAM)s);
		SendMessage(hwList, LB_SETCURSEL, isel_alloc, 0);
		SendMessage(hwList, WM_SETREDRAW, TRUE, 0);
		return;
	}
	void import_tilemap(){
		if (!(maps->data[isel].tilemap_type & 0x0020)) return;
		int format = SendDlgItemMessage(hWnd, IDC_COMBO0, CB_GETCURSEL, 0, 0);
		OPENFILENAME ofn;
		ZeroMemory(&ofn, sizeof ofn);
		char filename[MAX_PATH] = {0};
		char filetitle[64] = {0};
		char init_dir[MAX_PATH];
		strcpy_s(init_dir, sizeof init_dir, maps->project->filename);
		char *last_sep = strrchr(init_dir, '\\');
		*(last_sep + 1) = '\0';
		ofn.lStructSize			= sizeof ofn;
		ofn.hwndOwner			= hWnd;
		ofn.hInstance			= NULL;
		ofn.lpstrFilter			= tilemap_format_filter[format];
		ofn.lpstrFile			= filename;
		ofn.nMaxFile			= sizeof filename;
		ofn.lpstrFileTitle		= filetitle;
		ofn.nMaxFileTitle		= sizeof filetitle;
		ofn.lpstrInitialDir		= init_dir;
		ofn.Flags				= OFN_FILEMUSTEXIST;
		ofn.lpstrDefExt			= tilemap_format_ext[format];
		ofn.lpstrTitle			= "n`f[^̓ǂݍ";
		if (!GetOpenFileName(&ofn)) return;
		static bool (TilemapDialog::*import_function[])(LPCSTR, LPVOID, DWORD, DWORD*, bool) = {
			&TilemapDialog::import_compressed_tilemap,
			&TilemapDialog::import_expanded_tilemap,
			&TilemapDialog::import_sdk2_mapedit_ppj,
			&TilemapDialog::import_sdk2_mapedit_fmf,
			&TilemapDialog::import_map_conv_ppj,
			&TilemapDialog::import_map_conv_fmf,
		};
		BYTE *buf = new BYTE[32768];
		DWORD size;
		if (!(this->*import_function[format])(filename, buf, 65536, &size,
			(maps->data[isel].tilemap_type & 0x0400) != 0))
		{
			MessageBox(hWnd, "f[^̓ǂݍ݂Ɏs܂B", "G[", MB_OK | MB_ICONEXCLAMATION);
			delete[] buf;
			return;
		}
		DWORD size_max = maps->get_tilemap_max_size(isel);
		if (size > size_max){
			char s[1024];
			sprintf_s(s, sizeof s, "f[^̃TCY傫܂B\r\n"
				"ǂݍ񂾃f[^: %doCg\r\n"
				"ǂݍ݉\ȃTCY: %doCg", size, size_max);
			MessageBox(hWnd, s, "G[", MB_OK | MB_ICONEXCLAMATION);
			delete[] buf;
			return;
		}else{
			char s[1024];
			sprintf_s(s, sizeof s, "f[^̓ǂݍ݂ɐ܂B\r\n"
				"ǂݍ񂾃f[^: %doCg\r\n"
				"ǂݍ݉\ȃTCY: %doCg", size, size_max);
			MessageBox(hWnd, s, "n`f[^̓ǂݍ", MB_OK | MB_ICONINFORMATION);
		}
		if (tilemap_change[isel]){
			delete[] tilemap_change[isel];
		}else{
			data_changed();
		}
		tilemap_change[isel] = new BYTE[size_max];
		tilemap_size[isel] = size_max;
		memcpy_s(tilemap_change[isel], size_max,
			maps->project->get_pointer(maps->data[isel].tilemap_adrs),
			size_max);
		memcpy_s(tilemap_change[isel], size, buf, size);
		delete[] buf;
		return;
	}
	bool import_compressed_tilemap(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		HANDLE hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) return false;
		DWORD size_read;
		if (compressed){
			// kꂽf[^̂܂ܓǂݍށB
			*size = GetFileSize(hFile, NULL);
			if (*size > buf_size){
				CloseHandle(hFile);
				return false;
			}
			ReadFile(hFile, buf, *size, &size_read, NULL);
		}else{
			// kf[^WJēǂݍށB(܂ӖȂc)
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			DWORD size_file = GetFileSize(hFile, NULL);
			char *buf_file = new char[size_file];
			ReadFile(hFile, buf_file, size_file, &size_read, NULL);
			*size = ExpandData((unsigned char *)buf, buf_size, (const unsigned char *)buf_file);
			if (*size == -1){
				delete[] buf_file;
				CloseHandle(hFile);
				return false;
			}
			delete[] buf_file;
		}
		CloseHandle(hFile);
		return true;
	}
	bool import_expanded_tilemap(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		HANDLE hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) return false;
		DWORD size_read;
		if (compressed){
			// WJf[^kēǂݍށB
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			DWORD size_file = GetFileSize(hFile, NULL);
			char *buf_file = new char[size_file];
			ReadFile(hFile, buf_file, size_file, &size_read, NULL);
			*size = CompressData((unsigned char *)buf, buf_size, (const unsigned char *)buf_file, size_file);
			if (*size == -1){
				delete[] buf_file;
				CloseHandle(hFile);
				return false;
			}
			delete[] buf_file;
		}else{
			// WJf[^̂܂ܓǂݍށB
			*size = GetFileSize(hFile, NULL);
			if (*size > buf_size){
				CloseHandle(hFile);
				return false;
			}
			ReadFile(hFile, buf, *size, &size_read, NULL);
		}
		CloseHandle(hFile);
		return true;
	}
	bool import_sdk2_mapedit_ppj(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		extern bool InputPPJ_MapEdit(LPCSTR filename, WORD *buf, DWORD buf_size, DWORD *width, DWORD *height, bool bVert);
		bool bVert;
		int thickness;
		if (compressed){
			// PPJtϊk
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			BYTE *buf_inter = new BYTE[131072];
			DWORD width, height;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!InputPPJ_MapEdit(filename, (WORD *)buf_inter, 131072, &width, &height, bVert)){
				delete[] buf_inter;
				return false;
			}
			*size = CompressData((unsigned char *)buf, buf_size, (const unsigned char *)buf_inter, width*height*2);
			if (*size == -1){
				delete[] buf_inter;
				return false;
			}
			delete[] buf_inter;
		}else{
			// PPJtϊ
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			DWORD width, height;
			if (!InputPPJ_MapEdit(filename, (WORD *)buf, buf_size, &width, &height, bVert)){
				return false;
			}
		}
		return true;
	}
	bool import_sdk2_mapedit_fmf(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		extern bool InputFMF_MapEdit(LPCSTR filename, WORD *buf, DWORD buf_size, DWORD *width, DWORD *height, bool bVert);
		bool bVert;
		int thickness;
		if (compressed){
			// FMFtϊk
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			BYTE *buf_inter = new BYTE[131072];
			DWORD width, height;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!InputFMF_MapEdit(filename, (WORD *)buf_inter, 131072, &width, &height, bVert)){
				delete[] buf_inter;
				return false;
			}
			*size = CompressData((unsigned char *)buf, buf_size, (const unsigned char *)buf_inter, width*height*2);
			if (*size == -1){
				delete[] buf_inter;
				return false;
			}
			delete[] buf_inter;
		}else{
			// FMFtϊ
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			DWORD width, height;
			if (!InputFMF_MapEdit(filename, (WORD *)buf, buf_size, &width, &height, bVert)){
				return false;
			}
		}
		return true;
	}
	bool import_map_conv_ppj(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		extern bool InputPPJ_MapConv(LPCSTR filename, WORD *buf, DWORD buf_size, DWORD *width, DWORD *height, bool bVert);
		bool bVert;
		int thickness;
		if (compressed){
			// PPJtϊk
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			BYTE *buf_inter = new BYTE[131072];
			DWORD width, height;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!InputPPJ_MapConv(filename, (WORD *)buf_inter, 131072, &width, &height, bVert)){
				delete[] buf_inter;
				return false;
			}
			*size = CompressData((unsigned char *)buf, buf_size, (const unsigned char *)buf_inter, width*height*2);
			if (*size == -1){
				delete[] buf_inter;
				return false;
			}
			delete[] buf_inter;
		}else{
			// PPJtϊ
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			DWORD width, height;
			if (!InputPPJ_MapConv(filename, (WORD *)buf, buf_size, &width, &height, bVert)){
				return false;
			}
		}
		return true;
	}
	bool import_map_conv_fmf(LPCSTR filename, LPVOID buf, DWORD buf_size, DWORD *size, bool compressed){
		extern bool InputFMF_MapConv(LPCSTR filename, WORD *buf, DWORD buf_size, DWORD *width, DWORD *height, bool bVert);
		bool bVert;
		int thickness;
		if (compressed){
			// FMFtϊk
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			BYTE *buf_inter = new BYTE[131072];
			DWORD width, height;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!InputFMF_MapConv(filename, (WORD *)buf_inter, 131072, &width, &height, bVert)){
				delete[] buf_inter;
				return false;
			}
			*size = CompressData((unsigned char *)buf, buf_size, (const unsigned char *)buf_inter, width*height*2);
			if (*size == -1){
				delete[] buf_inter;
				return false;
			}
			delete[] buf_inter;
		}else{
			// FMFtϊ
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			DWORD width, height;
			if (!InputFMF_MapConv(filename, (WORD *)buf, buf_size, &width, &height, bVert)){
				return false;
			}
		}
		return true;
	}
	void export_tilemap(){
		if (!(maps->data[isel].tilemap_type & 0x0020)) return;
		int format = SendDlgItemMessage(hWnd, IDC_COMBO0, CB_GETCURSEL, 0, 0);
		OPENFILENAME ofn;
		ZeroMemory(&ofn, sizeof ofn);
		char filename[MAX_PATH] = {0};
		char filetitle[64] = {0};
		char init_dir[MAX_PATH];
		strcpy_s(init_dir, sizeof init_dir, maps->project->filename);
		char *last_sep = strrchr(init_dir, '\\');
		*(last_sep + 1) = '\0';
		ofn.lStructSize			= sizeof ofn;
		ofn.hwndOwner			= hWnd;
		ofn.hInstance			= NULL;
		ofn.lpstrFilter			= tilemap_format_filter[format];
		ofn.lpstrFile			= filename;
		ofn.nMaxFile			= sizeof filename;
		ofn.lpstrFileTitle		= filetitle;
		ofn.nMaxFileTitle		= sizeof filetitle;
		ofn.lpstrInitialDir		= init_dir;
		ofn.Flags				= OFN_OVERWRITEPROMPT | OFN_EXPLORER;
		ofn.lpstrDefExt			= tilemap_format_ext[format];
		ofn.lpstrTitle			= "n`f[^̏o";
		if (!GetSaveFileName(&ofn)) return;
		static bool (TilemapDialog::*export_function[])(LPCSTR, LPCVOID, bool) = {
			&TilemapDialog::export_compressed_tilemap,
			&TilemapDialog::export_expanded_tilemap,
			&TilemapDialog::export_sdk2_mapedit_ppj,
			&TilemapDialog::export_sdk2_mapedit_fmf,
			&TilemapDialog::export_map_conv_ppj,
			&TilemapDialog::export_map_conv_fmf,
		};
		if (!(this->*export_function[format])(filename,
			tilemap_change[isel]? (LPCVOID)tilemap_change[isel]
			: maps->project->get_pointer(maps->data[isel].tilemap_adrs),
			(maps->data[isel].tilemap_type & 0x0400) != 0))
		{
			MessageBox(hWnd, "f[^̏oɎs܂B", "G[", MB_OK | MB_ICONEXCLAMATION);
		}
		return;
	}

	bool export_compressed_tilemap(LPCSTR filename, LPCVOID data, bool src_compressed){
		HANDLE hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL,
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) return false;
		DWORD size_written;
		if (src_compressed){
			// kꂽf[^̂܂܏oB(Af[^TCY͐p̊֐ɋĂ炤)
			extern int ExtractDataSize(const unsigned char *buf);
			int size = ExtractDataSize((const unsigned char *)data);
			if (size == -1){
				CloseHandle(hFile);
				return false;
			}
			WriteFile(hFile, data, size, &size_written, NULL);
		}else{
			// k̃f[^kďoB(܂g͂Ȃc)
			extern int CompressData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src, unsigned int src_size);
			unsigned char *buf = new unsigned char[320];
			int size = CompressData(buf, 320, (const unsigned char *)data, 256);
			if (size == -1){
				delete[] buf;
				CloseHandle(hFile);
				return false;
			}
			WriteFile(hFile, buf, size, &size_written, NULL);
			delete[] buf;
		}
		CloseHandle(hFile);
		return true;
	}
	bool export_expanded_tilemap(LPCSTR filename, LPCVOID data, bool src_compressed){
		HANDLE hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL,
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) return false;
		DWORD size_written;
		if (src_compressed){
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			BYTE *buf = new BYTE[131072];
			int size = ExpandData((unsigned char *)buf, 131072, (const unsigned char *)data);
			if (size == -1){
				delete[] buf;
				CloseHandle(hFile);
				return false;
			}
			WriteFile(hFile, buf, size, &size_written, NULL);
			delete[] buf;
		}else{
			WriteFile(hFile, data, 256, &size_written, NULL);
		}
		CloseHandle(hFile);
		return true;
	}
	bool export_sdk2_mapedit_ppj(LPCSTR filename, LPCVOID data, bool src_compressed){
		extern bool OutputPPJ_MapEdit(LPCSTR filename, const WORD *buf_src, DWORD width, DWORD height, bool bVert, LPCSTR chip_filename);
		bool bVert;
		int thickness;
		char chipname[256];
		GetPrivateProfileString("MapIO", "PPJChipDirectory_MapEdit", "", chipname, 256, ConfigFileName);
		maps->get_chip_file_me(isel, chipname+strlen(chipname), 256-strlen(chipname));
		if (src_compressed){
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			BYTE *buf = new BYTE[131072];
			int size = ExpandData((unsigned char *)buf, 131072, (const unsigned char *)data);
			if (size == -1){
				delete[] buf;
				return false;
			}
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputPPJ_MapEdit(filename, (const WORD *)buf,
				bVert? thickness : size/2/thickness, bVert? size/2/thickness : thickness, bVert,
				chipname))
			{
				delete[] buf;
				return false;
			}
			delete[] buf;
		}else{
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputPPJ_MapEdit(filename, (const WORD *)data,
				bVert? thickness : 128/thickness, bVert? 128/thickness : thickness, bVert,
				chipname))
			{
				return false;
			}
		}
		return true;
	}
	bool export_sdk2_mapedit_fmf(LPCSTR filename, LPCVOID data, bool src_compressed){
		extern bool OutputFMF_MapEdit(LPCSTR filename, const WORD *buf_src, DWORD width, DWORD height, bool bVert);
		bool bVert;
		int thickness;
		if (src_compressed){
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			BYTE *buf = new BYTE[131072];
			int size = ExpandData((unsigned char *)buf, 131072, (const unsigned char *)data);
			if (size == -1){
				delete[] buf;
				return false;
			}
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputFMF_MapEdit(filename, (const WORD *)buf,
				bVert? thickness : size/2/thickness, bVert? size/2/thickness : thickness, bVert))
			{
				delete[] buf;
				return false;
			}
			delete[] buf;
		}else{
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputFMF_MapEdit(filename, (const WORD *)data,
				bVert? thickness : 128/thickness, bVert? 128/thickness : thickness, bVert))
			{
				return false;
			}
		}
		return true;
	}
	bool export_map_conv_ppj(LPCSTR filename, LPCVOID data, bool src_compressed){
		extern bool OutputPPJ_MapConv(LPCSTR filename, const WORD *buf_src, DWORD width, DWORD height, bool bVert, LPCSTR chip_filename);
		bool bVert;
		int thickness;
		char chipname[256];
		GetPrivateProfileString("MapIO", "PPJChipDirectory_MapConv", "chips\\", chipname, 256, ConfigFileName);
		maps->get_chip_file(isel, chipname+strlen(chipname), 256-strlen(chipname));
		if (src_compressed){
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			BYTE *buf = new BYTE[131072];
			int size = ExpandData((unsigned char *)buf, 131072, (const unsigned char *)data);
			if (size == -1){
				delete[] buf;
				return false;
			}
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputPPJ_MapConv(filename, (const WORD *)buf,
				bVert? thickness : size/2/thickness, bVert? size/2/thickness : thickness, bVert,
				chipname))
			{
				delete[] buf;
				return false;
			}
			delete[] buf;
		}else{
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputPPJ_MapConv(filename, (const WORD *)data,
				bVert? thickness : 128/thickness, bVert? 128/thickness : thickness, bVert,
				chipname))
			{
				return false;
			}
		}
		return true;
	}
	bool export_map_conv_fmf(LPCSTR filename, LPCVOID data, bool src_compressed){
		extern bool OutputFMF_MapConv(LPCSTR filename, const WORD *buf_src, DWORD width, DWORD height, bool bVert);
		bool bVert;
		int thickness;
		if (src_compressed){
			extern int ExpandData(unsigned char *buf, unsigned int buf_size, const unsigned char *buf_src);
			BYTE *buf = new BYTE[131072];
			int size = ExpandData((unsigned char *)buf, 131072, (const unsigned char *)data);
			if (size == -1){
				delete[] buf;
				return false;
			}
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputFMF_MapConv(filename, (const WORD *)buf,
				bVert? thickness : size/2/thickness, bVert? size/2/thickness : thickness, bVert))
			{
				delete[] buf;
				return false;
			}
			delete[] buf;
		}else{
			bool bVert;
			int thickness;
			maps->get_tilemap_attribute(isel, &bVert, &thickness);
			if (!OutputFMF_MapConv(filename, (const WORD *)data,
				bVert? thickness : 128/thickness, bVert? 128/thickness : thickness, bVert))
			{
				return false;
			}
		}
		return true;
	}

	void save_data(){
		Tilemap_UndoMethod *undo = new Tilemap_UndoMethod(maps);
		bool bChange = false;
		for (int i=0; i<maps->map_count; i++){
			if (alloc_change[i]){
				MapAllocationData alloc_before;
				alloc_before.load(maps->project,
					maps->pAllocationAdrs.bank, maps->data[i].allocation_adrs);
				undo->add_alloc_change(i, alloc_before, *alloc_change[i]);
				alloc_change[i]->store(maps->project,
					maps->pAllocationAdrs.bank, maps->data[i].allocation_adrs);
				bChange = true;
			}
			if (tilemap_change[i]){
				undo->add_tilemap_change(i,
					(const BYTE *)maps->project->get_pointer(maps->data[i].tilemap_adrs),
					(const BYTE *)tilemap_change[i], tilemap_size[i]);
				maps->project->write_data(tilemap_change[i], tilemap_size[i],
					maps->data[i].tilemap_adrs);
				bChange = true;
			}
		}
		if (bChange){
			maps->project->undo_buffer.add_method(undo);
			maps->project->levels.redraw_all();
		}else{
			delete undo;
		}
	}
	void end(WORD code){
		char s[256];
		sprintf_s(s, sizeof s, "%d", SendDlgItemMessage(hWnd, IDC_COMBO0, CB_GETCURSEL, 0, 0));
		WritePrivateProfileString("MapIO", "DefaultFormat", s, ConfigFileName);
		if (code == IDOK) save_data();
		EndDialog(hWnd, (INT_PTR)code);
	}
public:
	TilemapDialog(MapsManager *_maps) : maps(_maps), alloc_change(NULL), tilemap_change(NULL), tilemap_size(NULL){
		alloc_change = new MapAllocationData*[maps->map_count];
		tilemap_change = new BYTE*[maps->map_count];
		tilemap_size = new DWORD[maps->map_count];
		for (int i=0; i<maps->map_count; i++){
			alloc_change[i] = NULL;
			tilemap_change[i] = NULL;
			tilemap_size[i] = 0;
		}
	}
	~TilemapDialog(){
		if (alloc_change){
			for (int i=0; i<maps->map_count; i++){
				if (alloc_change[i]) delete alloc_change[i];
			}
			delete[] alloc_change;
		}
		if (tilemap_change){
			for (int i=0; i<maps->map_count; i++){
				if (tilemap_change[i]) delete[] tilemap_change[i];
			}
			delete[] tilemap_change;
		}
		if (tilemap_size){
			delete[] tilemap_size;
		}
	}

};





int MapsManager::tilemap_edit_dialog(HWND hw_owner, int init_map_num, int init_map_index){
	TilemapDialog dialog(this);
	return dialog.dialog_box(hw_owner, (DWORD)((WORD)init_map_num) | (DWORD)((WORD)init_map_index << 16));
}







